"""
# logutils.py - Utilities for logging etc
#
#####################################################
# OVERVIEW:
#
# Logging utilities
#
######################################################
"""

import sys
import re
import os
import time
import stat
import logging

def python3():
    return sys.version_info[0] >= 3

import traceback

def notify_users(mail_form_url, subject, msg):
    "Uses simple form to send an email to fixed set of users"
    if not mail_form_url:
        return
    response = None
    try:
        if python3():
            import urllib.parse
            import urllib.request
            data = urllib.parse.urlencode({'subject': subject, 'message': msg})
            response = urllib.request.urlopen(mail_form_url, data.encode('ascii'))
        else:
            import urllib
            data = urllib.urlencode({'subject': subject, 'message':msg})
            response = urllib.urlopen(mail_form_url, data)
    except Exception as e:
        print("Failed to notify:", str(e))
        sys.stdout.flush()
    return response

def save_existing_file(file_name):
    "Ensures we don't overwrite existing files by renaming them"
    if os.path.exists(file_name):
        new_name = os.path.basename(file_name)
        (new_name, ext) = new_name.rsplit(".", 1)
        new_name = "%s-%s.%s" % (new_name,
                    time.strftime("%Y%m%d%H%M%S", time.localtime()), ext)
        os.rename(file_name, new_name)
        return new_name
    return None

def get_unique_file_name(file_path):
    "Ensures we have a unique file name"
    i = 1
    ldir = os.path.dirname(file_path)
    (base_file_name, ext) = os.path.basename(file_path).rsplit(".", 1)
    while os.path.exists(file_path):
        file_path = os.path.join(ldir, "%s-%d.%s" % (base_file_name, i, ext))
        i += 1
    return file_path

def get_log_file_name():
    prefix = sys.argv[0].split(".py")[0]
    return get_unique_file_name(os.path.join(os.getcwd(),
                            "log-%s-%s.log" % (os.path.basename(prefix),
                            time.strftime('%Y%m%d%H%M%S', time.localtime()))))

class Logger(object):
    "Do logging"

    def __init__(self, mail_form_url=None):
        self.mail_form_url = mail_form_url
        prefix = sys.argv[0].split(".py")[0]
        self.logfile_name = get_unique_file_name(os.path.join(os.getcwd(),
                                "log-%s-%s.log" % (os.path.basename(prefix),
                                time.strftime('%Y%m%d%H%M%S', time.localtime()))))
        if python3():
            # Handle any unicode errors
            self.logfile = open(self.logfile_name, "w", errors='backslashreplace')
        else:
            self.logfile = open(self.logfile_name, "w")
        self.time_last_notified = time.time()
        self.saved_output = []
        self.saved_log = []
        self.output("Logging to: %s" % self.logfile_name)

    def shutdown(self):
        self.logfile.close()

    def save_log(self, line):
        "Save to circular buffer"
        self.saved_log.append(line)
        if len(self.saved_log) > 100:
            del self.saved_log[0]

    def save_output(self, line):
        "Save to circular buffer"
        self.saved_output.append(line)
        if len(self.saved_output) > 50:
            del self.saved_output[0]

    def log(self, *params):
        "Write to log file"
        line = ", ".join([str(x) for x in params])
        self.logfile.write(line)
        self.logfile.write("\n")
        self.logfile.flush()
        self.save_log(line)

    def output(self, *params):
        "Print to stdout and log"
        timestamp = time.strftime('%Y/%m/%d %H:%M:%S', time.localtime())
        self.log(timestamp, params)
        msg = "%s: %s" % (timestamp, ", ".join([str(x) for x in params]))
        line = msg.encode(errors='backslashreplace')
        print(line)
        sys.stdout.flush()
        self.save_output(line)
        if time.time() - self.time_last_notified > 30 * 60: # 30 mins
            self.time_last_notified = time.time()
            notify_users(self.mail_form_url, "Regular update for %s" % sys.argv[0], "\n".join(self.saved_output))

    def report_exception(self, msg=None):
        "Notify users that we have had a problem"
        msg = "Exception caught:\n" + traceback.format_exc()
        for line in msg.split("\n"):
            self.output(line)
        notify_users(self.mail_form_url, "Exception in %s" % sys.argv[0], "\n".join(self.saved_log))

    def notify(self, subject, body):
        "Notify users of a message"
        return notify_users(self.mail_form_url, "%s %s" % (subject, sys.argv[0]), 
                     body + "\n".join([str(x) for x in self.saved_output]))

class ArgLogRecord(logging.LogRecord):
    """Custom formatting - just prints out any arguments passed"""
    def __init__(self, name, level, pathname, lineno,
                 msg, args, exc_info, func=None):
        logging.LogRecord.__init__(self, name, level, pathname, lineno,
                 msg, args, exc_info, func)

    def getMessage(self):
        """
        Return the message for this LogRecord.

        Return the message for this LogRecord after merging any user-supplied
        arguments with the message.
        """
        msg = str(self.msg)
        if self.args:
            try:
                msg = msg % self.args
            except TypeError as ex:
                msg += ", ".join([str(x) for x in self.args])
        return msg

class ArgLogger(logging.getLoggerClass()):
    "Register our logging class"
    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func, extra):
        return ArgLogRecord(name, level, fn, lno, msg, args, exc_info, func)

def getLogger(logger_name):
    "Register our logger and initialise everything"
    logging.setLoggerClass(ArgLogger)
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    fh = logging.FileHandler(filename=get_log_file_name())
    fh.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    fh.setFormatter(formatter)
    logger.addHandler(ch)
    logger.addHandler(fh)
    return logger

def test():
    "Test the above"
    logger = InitLogging('testlogger')
    logger.debug('debug message')
    logger.info('info message')
    logger.warn('warn message')
    logger.error('error message')
    logger.critical('critical message')

    logger.info("Some", "text", "params")
    logger.info([1, 2, "three", 1.2], "more")
    logger.info("Unicode text", '\ufffd')
    logging.shutdown()

if __name__ == "__main__":
    test()
